library kvs;

import 'dart:io';
import 'dart:convert';
import 'package:flutter/services.dart';
import 'package:flutter/widgets.dart';
import 'package:path_provider/path_provider.dart';
import 'cipher.dart';

class KVS<K, V> {
  String name;
  String? documentDirectory;
  late Map<K, V> data;
  int version = 0;
  bool writeLock = false;
  Cipher? cipher;

  KVS({
    this.name = 'default',
    this.documentDirectory,
    this.cipher,
  });

  static Future<KVS<K, V>> init<K, V>({
    String name = 'default',
    String? documentDirectory,
    Cipher? cipher,
    Map<K, V>? initialData,
  }) async {
    Map<K, V> data = {...(initialData ?? {})};

    WidgetsFlutterBinding.ensureInitialized();

    documentDirectory ??= (await getApplicationDocumentsDirectory()).path;

    var file = File('$documentDirectory/kvs-$name.json');

    if (file.existsSync()) {
      String fileAsString;

      if (cipher != null) {
        var encryptedBytes = await file.readAsBytes();
        var decryptedBytes = await cipher.decrypt(encryptedBytes);

        fileAsString = utf8.decode(decryptedBytes);
      } else {
        fileAsString = await file.readAsString();
      }

      try {
        data = jsonDecode(fileAsString);
      } catch (error) {
        data = {};
      }
    } else {
      data = {};
    }

    var kvs = KVS<K, V>(
      name: name,
      documentDirectory: documentDirectory,
      cipher: cipher,
    );

    kvs.data = data;

    return kvs;
  }

  flush() async {
    if (writeLock) return;

    writeLock = true;
    int writtenVersion = version;
    var file = File('$documentDirectory/kvs-$name.json');

    String fileAsString = jsonEncode(data);

    if (cipher != null) {
      var plainBytes = Uint8List.fromList(utf8.encode(fileAsString));
      var encryptedBytes = await cipher!.encrypt(plainBytes);

      await file.writeAsBytes(encryptedBytes, flush: true);
    } else {
      await file.writeAsString(fileAsString, flush: true);
    }

    writeLock = false;

    // if version changed in meantime flush it
    if (writtenVersion < version) {
      flush();
    }

    return true;
  }

  V? operator [](K key) => data[key];

  void operator []=(K key, V value) {
    data[key] = value;
    version++;

    flush();
  }

  void addAll(Map<K, V> items) {
    data.addAll(items);
    flush();
  }

  Map<RK, RV> cast<RK, RV>() => data.cast<RK, RV>();

  void clear() {
    data.clear();
    flush();
  }

  bool containsKey(K key) => data.containsKey(key);
  bool containsValue(V value) => data.containsValue(value);
  void forEach(void Function(K, V) action) => data.forEach(action);
  Iterable<K> get keys => data.keys;
  int get length => data.length;
  Map<K2, V2> map<K2, V2>(MapEntry<K2, V2> Function(K, V) convert) =>
      data.map(convert);
  Iterable<V> get values => data.values;
}
